/* empty - run processes under pseudo-terminal sessions
 * 
 * Copyright (C) 2005 Mikhail E. Zakharov
 * empty was written by Mikhail E. Zakharov. This software was based on the 
 * basic idea of pty version 4.0 Copyright (c) 1992, Daniel J. Bernstein, but
 * no code was ported from pty4.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *  
 *  1. Redistributions of source code must retain the above copyright
 *     notice immediately at the beginning of the file, without modification,
 *     this list of conditions, and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <unistd.h>
#include <sys/types.h>
#ifdef svr4
	#include <stropts.h>
#endif
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <termios.h>
#ifndef svr4
	#include <err.h>
#endif
#include <errno.h>
#include <sys/select.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <syslog.h>
#include <stdarg.h>
#include <sys/param.h>

#define program "empty"
#define version "0.6.0b"

#define tmpdir "/tmp"

/* -------------------------------------------------------------------------- */
static void usage(void);
int toint(char *intstr);
void wait4child(int child, char *argv0);
int mfifo(char *fname, int mode);
void perrx(int ex_code, const char *err_text);
void perrxslog(int ex_code, const char *err_text);
int pidbyppid(int ppid, int lflg);
void clean(void);
void fsignal(int sig);
int watch4str(int ifd, int ofd, char *get, char *resp,
		int nflg, int vflg, int cflg, int timeout);

/* -------------------------------------------------------------------------- */
int     master, slave;
int	child;
int	pid;
char	*in = NULL, *out = NULL;
int	ifd, ofd;
int	status;
char	buf[BUFSIZ];
fd_set	rfd;
char	*argv0 = NULL;
/* -------------------------------------------------------------------------- */
int main (int argc, char *argv[]) {
	struct	winsize win;
	struct	termios tt;
	int	i, fdmax, cc, n, ch;
	int	fflg = 0;		/* spawn, fork */
	int	wflg = 0;		/* watch for string [respond] */
	int	sflg = 0;		/* send */
	int	kflg = 0;		/* kill */
	int	lflg = 0;		/* list your jobs */
	int	iflg = 0;		/* in */
	int	oflg = 0;		/* out */
	int	nflg = 0;		/* newline */
	int	cflg = 0;		/* use stdin instead of FIFO */
	int	vflg = 0;		/* kvazi verbose mode OFF */
	int	timeout = 10;		/* wait N secs for the responce */
	
	int	kpid = -1;
	int	ksig = SIGTERM;

	char    *get = NULL, *resp = NULL;

	int	ppid;			/* Shell PID */
	char	infifo[MAXPATHLEN];
	char	outfifo[MAXPATHLEN];

#ifdef svr4
	char	*slave_name;
	int	pgrp;
#endif

	while ((ch = getopt(argc, argv, "fk:wsli:o:r:t:ncvh")) != -1)
		switch (ch) {
			case 'f':
				fflg = 1;
				break;
			case 'k':
				/* Send signal */
				if ((kpid = toint(optarg)) < 0) {
					fprintf(stderr, 
						"Fatal: wrong -k value\n");
					(void)usage();
				}

				kflg = 1;
				break;
			case 'w':
				wflg = 1;
				break;
			case 's':
				sflg = 1;
				break;
			case 'l':
				lflg = 1;
				break;
			case 'i': 
				in = optarg;
				iflg = 1;
				break;
			case 'o': 
				out = optarg;
				oflg = 1;
				break;
			case 't':
				/* wait N secs for the responce. Use with -w */
				if ((timeout = toint(optarg)) < 1) {
					fprintf(stderr,
						"Fatal: wrong -t value\n");
					(void)usage();
				}
				break;
			case 'c':
				/* use stdin instead of FIFO */
				cflg = 1;
				break;
		
			case 'n':
				/* do NOT append newline */
				nflg = 1;
				break;
			case 'v':
				vflg = 1;
				break;
			case 'h':
			default:
				(void)usage();
		}
	argc -= optind;
	argv += optind;

	
	if ((fflg + kflg + wflg + sflg + lflg) != 1)
		(void)usage();

	ppid = getppid();
	
	if (kflg) {	/* kill PID with the SIGNAL */
		if (argv[0])
			ksig = toint(argv[0]);
		if (kill(kpid, ksig) == -1)
			(void)perrx(1, "Can't kill()");

		exit(0);
	}

	if (lflg) {
		pidbyppid(ppid, lflg);

		exit(0);
	}

	if (sflg) {	/* we want to send data */
		pid = pidbyppid(ppid, lflg); 
		sprintf(outfifo, "%s/%s.%d.%d.in", tmpdir, program, ppid, pid);
		out = (char  *)outfifo;

		if ((ofd = open(out, O_WRONLY)) == -1)
			(void)perrx(1, "Fatal open(out, ...)");
		
		if (!cflg && argv[0] != NULL) {
			if (write(ofd, argv[0], strlen(argv[0])) == -1)
				(void)perrx(1, "Fatal write(ofd, cmdl_buf, ...)");
		} else
			while ((cc  = read(0, buf, sizeof(buf))) > 0) {
				if (cc == -1)
					(void)perrx(1, "Fatal read(stdin, stdin_buf, ...)");
				if (write(ofd, buf, cc) == -1)
					(void)perrx(1, "Fatal write(ofd, stdin_buf, ...)");
			}

		if (!nflg) /* append the '\n' */
			if (write(ofd, "\n", 1) == -1)
				(void)perrx(1, "Fatal write(ofd, newline, ...)");
		
		exit(0);
	}

	if (argc == 0)
		(void)usage();

	if (wflg) {	/* watch 4 string [send response] */
		switch (argc) {
			case 2:
				resp = argv[1];
			case 1:
				get = argv[0];
		}
	
		pid = pidbyppid(ppid, lflg); 
		sprintf(infifo, "%s/%s.%d.%d.out", tmpdir, program, ppid, pid);
		sprintf(outfifo, "%s/%s.%d.%d.in", tmpdir, program, ppid, pid);
		in = (char *)infifo;
		out = (char *)outfifo;

		if ((ifd = open(in, O_RDONLY)) == -1)
			(void)perrx(1, "Fatal open(in, ...)");
		if ((ofd = open(out, O_WRONLY)) == -1)
			(void)perrx(1, "Fatal open(out, ...)");
	
		exit(watch4str(ifd, ofd, get, resp, nflg, vflg, cflg, timeout));
	}

	(void)openlog(program, 0, 0);
	(void)syslog(LOG_NOTICE, "version %s started", version);
	argv0 = argv[0];
	
	(void)tcgetattr(STDIN_FILENO, &tt);
        (void)ioctl(STDIN_FILENO, TIOCGWINSZ, &win);

	if ((pid = fork()) == -1)
		(void)perrxslog(1, "Fatal fork()");
	if (pid > 0) exit(0);
	
	if (setsid() < 0)
		(void)perrxslog(1, "Fatal setsid()");
	
	if ((pid = fork()) == -1)
		(void)perrxslog(1, "Fatal fork()");
	if (pid > 0) exit(0);
	for (i = 1; i < 32; i++) signal(i, fsignal);
	
	if (!(iflg && oflg)) {
		pid = getpid();						/* MY PID */
		sprintf(infifo, "%s/%s.%d.%d.in", tmpdir, program, ppid, pid);
		sprintf(outfifo, "%s/%s.%d.%d.out", tmpdir, program, ppid, pid);
		in = (char *)infifo;
		out = (char *)outfifo;
	}
	
#ifndef OpenBSD
	if ((ifd = mfifo(in, O_RDWR)) == -1)
		(void)perrxslog(1, "Fatal mfifo(in, ...)");
	if ((ofd = mfifo(out, O_RDWR)) == -1)
		(void)perrxslog(1, "Fatal mfifo(out, ...");
#else
	if ((ifd = mfifo(in, O_RDONLY|O_NONBLOCK)) == -1)
		(void)perrxslog(1, "Fatal mfifo(in, ...)");
	if ((ofd = mfifo(out, O_WRONLY)) == -1)
		(void)perrxslog(1, "Fatal mfifo(out, ...");
#endif

	
#ifndef svr4
	if (openpty(&master, &slave, NULL, &tt, &win) == -1)
		(void)perrxslog(1, "Fatal openpty()");
#else
	if ((master = open("/dev/ptmx", O_RDWR)) == -1)
		(void)perrxslog(1, "Fatal open(\"/dev/ptmx\"), ...");

	if (unlockpt(master) == -1)
		(void)perrxslog(1, "Fatal unlockpt(master)");

	if ((slave_name = (char *)ptsname(master)) == NULL)
		(void)perrxslog(1, "Fatal ptsname(master)");
#endif
		
	if ((child = fork()) < 0) {
		(void)clean();
		(void)perrxslog(1, "Fatal fork()");
	}
	
	if (child == 0) {
		(void)close(master);
#ifndef svr4
		login_tty(slave);
		cfmakeraw(&tt);
#else
		if ((pgrp = setsid()) == -1)
			(void)syslog(LOG_NOTICE, "Can't setsid(): %m");
		
		if ((slave = open(slave_name, O_RDWR)) == -1)
			(void)perrxslog(1, "Fatal open(slave_name, ...)");

		ioctl(slave, I_PUSH, "ptem");
		ioctl(slave, I_PUSH, "ldterm");
		ioctl(slave, I_PUSH, "ttcompat");
		
		dup2(slave, 0);
	       	dup2(slave, 1);
	       	dup2(slave, 2);

		if (tcsetpgrp(0, pgrp) == -1)
			(void)perrxslog(1, "Fatal tcsetpgrp()");
#endif

		tt.c_lflag &= ~ECHO;
		(void)tcsetattr(STDIN_FILENO, TCSAFLUSH, &tt);

		(void)syslog(LOG_NOTICE, "%s forked", argv[0]);
		execvp(argv[0], argv);
		(void)syslog(LOG_NOTICE, "Failed loading %s: %m", argv[0]);

		(void)clean();
		(void)kill(0, SIGTERM);
		exit(0);
	}

	FD_ZERO(&rfd);	
	for (;;) {
		FD_SET(master, &rfd);
		FD_SET(ifd, &rfd);
	
		n = select(master + 1, &rfd, 0, 0, NULL);
		if (n > 0 || errno == EINTR) {
			if (FD_ISSET(ifd, &rfd)) 
				if ((cc = read(ifd, buf, sizeof(buf))) > 0)
					(void)write(master, buf, cc);

			if (FD_ISSET(master, &rfd))
				if ((cc = read(master, buf, sizeof(buf))) > 0)
					(void)write(ofd, buf, cc);
		}
	}
/* 	This never reached */
/*	return 0; */
}

/* -------------------------------------------------------------------------- */
static void usage(void) {
	(void)fprintf(stderr,
"%s-%s\n\
usage:\tempty -f [-i fifo1 -o fifo2] command [command args]\n\
\tempty -w [-cnv] [-t n] [-i fifo2 -o fifo1] keyphrase [response]\n\
\tempty -s [-cn] [-o fifo1] [request]\n\
\tempty -l\n\
\tempty -k pid [signal]\n", program, version);
	exit(1);
}

/* -------------------------------------------------------------------------- */
int toint(char *intstr) {
	int	in;

	in = strtol(intstr, (char **)NULL, 10);
	if (in == 0 && errno == EINVAL) {
		fprintf(stderr, "Wrong integer value: %s\n", intstr);
		usage();
	}
	return in;
}

/* -------------------------------------------------------------------------- */
int pidbyppid(int ppid, int lflg) {
	char fmask[MAXPATHLEN];
	char fname[MAXPATHLEN];
	const char *sep = ".";
	DIR *dir;
	struct dirent *dent;
	int len;
	char *chpid, *tail;
	int pid, maxpid;
	int header = 1;
	
	/* form this line: empty.ppid */
	sprintf(fmask, "%s%s%d", program, sep, ppid);
	len = strlen(fmask);

	if ((dir = opendir(tmpdir)) == NULL)
		(void)perrx(1, "Cant opendir()");
	
	while ((dent = readdir(dir)) != NULL) {
		if (!strncmp(fmask, dent->d_name, len)) {
			strcpy(fname, dent->d_name);

			strtok(fname, sep);		/* empty */
			strtok(NULL, sep);		/* PPID */
			chpid = strtok(NULL, sep);	/* PID */
			tail = strtok(NULL, sep);	/* IN or OUT */

			if (chpid != NULL) {
				pid = toint(chpid);
				if (pid > maxpid)
					maxpid = pid;
			}

			if (lflg) {
				if (header) {
					printf("PPID\tPID\tTYPE\tFILENAME\n");
					header --;
				}
					
				printf("%d\t%d\t%s\t%s/%s\n",
					ppid, pid, tail, tmpdir, dent->d_name);
			}
		}
	}

	closedir(dir);
	
	if (lflg && pid)
		printf("\n%d\t%d\tcurrent\n", ppid, maxpid);
	
	return maxpid;
}

/* -------------------------------------------------------------------------- */
void wait4child(int child, char *argv0) {
	while ((pid = wait3(&status, WNOHANG, 0)) > 0)
		if (pid == child) 
			(void)syslog(LOG_NOTICE, "%s exited", argv0);
}	

/* -------------------------------------------------------------------------- */
int mfifo(char *fname, int mode) {
	if (mkfifo(fname, S_IFIFO|S_IRWXU) == -1)
		return -1;
	
	return open(fname, mode);
}

/* -------------------------------------------------------------------------- */
void clean(void) {
	(void)close(master);
       	(void)close(ifd);
       	(void)close(ofd);
	(void)unlink(in);
	(void)unlink(out);
}		

/* -------------------------------------------------------------------------- */
void perrx(int ex_code, const char *err_text) {
	(void)perror(err_text);
	exit(ex_code);
}

/* -------------------------------------------------------------------------- */
void perrxslog(int ex_code, const char *err_text) {
	(void)syslog(LOG_NOTICE, "%s", err_text);
	(void)closelog();
	exit(ex_code);
}

/* -------------------------------------------------------------------------- */
void fsignal(int sig) {
	switch(sig) {
		case SIGTERM:
		case SIGINT:
		case SIGQUIT:
		case SIGSEGV:
			break;
		case SIGCHLD:
			wait4child(child, argv0);
			(void)syslog(LOG_NOTICE,
				"version %s finished", version);
	}

	(void)clean();
	(void)closelog();
	exit(0);
}

/* -------------------------------------------------------------------------- */
int watch4str(int ifd, int ofd, char *get, char *resp, 
		int nflg, int vflg, int cflg, int timeout) {

	int	n, cc, ccc;
	time_t	stime, ntime;
	struct	timeval tv;
	char    cbuf[BUFSIZ];

	stime = time(0);
	tv.tv_sec = timeout;
	tv.tv_usec = 0;
	
	FD_ZERO(&rfd);
	for (;;) {
		FD_SET(ifd, &rfd);
		n = select(ifd + 1, &rfd, 0, 0, &tv);
		if (n < 0 && errno != EINTR) 
			perrx(1, "Fatal select()");

		if (FD_ISSET(ifd, &rfd)) {
			if ((cc = read(ifd, buf, sizeof(buf)-1)) > 0) {
				buf[cc] = '\0';
				if (strstr(buf, get) != NULL) {
					/* Got WHAT we EXPECTED */
					if (vflg)
						(void)printf("%s", buf);

					if (cflg) {
						while ((ccc  = read(0, cbuf, sizeof(cbuf))) > 0) {
							if (ccc == -1)
								(void)perrx(1,
								"Fatal read(stdin, stdin_buf ...)");
							if (write(ofd, cbuf, ccc) == -1)
								(void)perrx(1,
								"Fatal write(ofd, stdin_buf ...)");
						}
					} else 
						if (resp) 
							write(ofd, resp,
								strlen(resp));
				
					if (!nflg && (cflg || resp)) 
						write(ofd, "\n", 1);
					
					return 0;
				}

				memmove(buf, buf + cc - strlen(get) + 1,
					strlen(get) - 1);
			}

			if (cc <= 0) {
				/* Got EOF or ERROR */
				if (vflg)
					(void)fprintf(stderr,
						"%s: Got nothing in output\n",
						program);
				return 1;
			}
		}

		ntime = time(0);
		if ((ntime - stime) >= timeout) {
			(void)fprintf(stderr,
				"%s: Keyphrase wasn't found. Exit on timeout\n",
				program);
			exit(-1);
		}
	}
}
